@@ -6,23 +6,21 @@ from email.header import Header
 from email.mime.application import MIMEApplication
 from email.mime.multipart import MIMEMultipart
 from email.mime.nonmultipart import MIMENonMultipart
-from email.utils import formataddr, getaddresses, parseaddr
+from email.utils import formataddr, getaddresses
 
-from sql import Null
-from sql.operators import Concat
 from genshi.template import TextTemplate
 
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, fields
+from trytond.model import ModelSQL, ModelView, fields
 from trytond.pool import Pool
-from trytond.pyson import Eval
+from trytond.pyson import Eval, TimeDelta
 from trytond.report import get_email
-from trytond.sendmail import sendmail_transactional, SMTPDataManager
+from trytond.sendmail import SMTPDataManager, sendmail_transactional
+from trytond.tools.email_ import convert_ascii_email, set_from_header
 from trytond.transaction import Transaction
-from .exceptions import TemplateError
 
-from trytond.ir.resource import ResourceAccessMixin
+from .exceptions import TemplateError
 
 
 class Email(ModelSQL, ModelView):
@@ -42,7 +40,6 @@ class Email(ModelSQL, ModelView):
         domain=[
             ('model.model', '=', Eval('model')),
             ],
-        depends=['model'],
         help="The field that contains the recipient(s).")
     fallback_recipients = fields.Many2One(
         'res.user', "Recipients Fallback User",
@@ -52,14 +49,12 @@ class Email(ModelSQL, ModelView):
         states={
             'invisible': ~Eval('recipients'),
             },
-        depends=['recipients'],
         help="User notified when no recipients e-mail is found")
     recipients_secondary = fields.Many2One(
         'ir.model.field', "Secondary Recipients",
         domain=[
             ('model.model', '=', Eval('model')),
             ],
-        depends=['model'],
         help="The field that contains the secondary recipient(s).")
     fallback_recipients_secondary = fields.Many2One(
         'res.user', "Secondary Recipients Fallback User",
@@ -69,14 +64,12 @@ class Email(ModelSQL, ModelView):
         states={
             'invisible': ~Eval('recipients_secondary'),
             },
-        depends=['recipients'],
         help="User notified when no secondary recipients e-mail is found")
     recipients_hidden = fields.Many2One(
         'ir.model.field', "Hidden Recipients",
         domain=[
             ('model.model', '=', Eval('model')),
             ],
-        depends=['model'],
         help="The field that contains the hidden recipient(s).")
     fallback_recipients_hidden = fields.Many2One(
         'res.user', "Hidden Recipients Fallback User",
@@ -86,7 +79,6 @@ class Email(ModelSQL, ModelView):
         states={
             'invisible': ~Eval('recipients_hidden'),
             },
-        depends=['recipients_hidden'],
         help="User notified when no hidden recipients e-mail is found")
 
     contact_mechanism = fields.Selection(
@@ -103,14 +95,20 @@ class Email(ModelSQL, ModelView):
         domain=[
             ('model', '=', Eval('model')),
             ],
-        depends=['model'],
         help="The reports used as attachments.")
 
     triggers = fields.One2Many(
         'ir.trigger', 'notification_email', "Triggers",
         domain=[('model.model', '=', Eval('model'))],
-        depends=['model'],
         help="Add a trigger for the notification.")
+    send_after = fields.TimeDelta(
+        "Send After",
+        domain=['OR',
+            ('send_after', '=', None),
+            ('send_after', '>=', TimeDelta()),
+            ],
+        help="The delay after which the email must be sent.\n"
+        "Applied if a worker queue is activated.")
 
     model = fields.Function(
         fields.Char("Model"), 'on_change_with_model', searcher='search_model')
@@ -171,13 +169,14 @@ class Email(ModelSQL, ModelView):
         with Transaction().set_context(usage=self.contact_mechanism):
             return EmailTemplate.get_languages(value)
 
-    def get_email(self, record, from_, to, cc, bcc, languages):
+    def get_email(self, record, sender, to, cc, bcc, languages):
         pool = Pool()
         Attachment = pool.get('notification.email.attachment')
 
         # TODO order languages to get default as last one for title
         content, title = get_email(self.content, record, languages)
         language = list(languages)[-1]
+        from_ = sender
         with Transaction().set_context(language=language.code):
             notification = self.__class__(self.id)
             if notification.from_:
@@ -195,105 +194,143 @@ class Email(ModelSQL, ModelView):
         else:
             msg = content
 
-        msg['From'] = from_
-        msg['To'] = ', '.join(formataddr(parseaddr(a)) for a in to)
-        msg['Cc'] = ', '.join(formataddr(parseaddr(a)) for a in cc)
+        set_from_header(msg, sender, from_)
+        msg['To'] = ', '.join(
+            formataddr((n, convert_ascii_email(a)))
+            for n, a in getaddresses(to))
+        msg['Cc'] = ', '.join(
+            formataddr((n, convert_ascii_email(a)))
+            for n, a in getaddresses(cc))
         msg['Subject'] = Header(title, 'utf-8')
         msg['Auto-Submitted'] = 'auto-generated'
-        return msg
-
-    def get_log(self, record, trigger, msg, bcc=None):
-        return {
-            'recipients': msg['To'],
-            'recipients_secondary': msg['Cc'],
-            'recipients_hidden': bcc,
-            'resource': str(record),
-            'notification': trigger.notification_email.id,
-            'trigger': trigger.id,
-            }
+        return msg, title
 
     @classmethod
     def trigger(cls, records, trigger):
         "Action function for the triggers"
-        if not trigger.notification_email:
+        notification_email = trigger.notification_email
+        if not notification_email:
             raise ValueError(
                 'Trigger "%s" is not related to any email notification'
                 % trigger.rec_name)
-        trigger.notification_email.send_email(records, trigger)
+        if notification_email.send_after:
+            with Transaction().set_context(
+                    queue_scheduled_at=trigger.notification_email.send_after):
+                notification_email.__class__.__queue__._send_email_queued(
+                    notification_email, [r.id for r in records], trigger.id)
+        else:
+            notification_email.send_email(records, trigger)
+
+    def _send_email_queued(self, ids, trigger_id):
+        pool = Pool()
+        Model = pool.get(self.model)
+        Trigger = pool.get('ir.trigger')
+        records = Model.browse(ids)
+        trigger = Trigger(trigger_id)
+        self.send_email(records, trigger)
 
     def send_email(self, records, trigger):
         pool = Pool()
-        Log = pool.get('notification.email.log')
+        Email = pool.get('ir.email')
         datamanager = SMTPDataManager()
         Transaction().join(datamanager)
         from_ = (config.get('notification_email', 'from')
             or config.get('email', 'from'))
-        logs = []
+        emails = []
         for record in records:
-            languagues = set()
-            to = []
-            if self.recipients:
-                recipients = getattr(record, self.recipients.name, None)
-                if recipients:
-                    languagues.update(self._get_languages(recipients))
-                    to = self._get_addresses(recipients)
-            if not to and self.fallback_recipients:
-                languagues.update(
-                    self._get_languages(self.fallback_recipients))
-                to = self._get_addresses(self.fallback_recipients)
-
-            cc = []
-            if self.recipients_secondary:
-                recipients_secondary = getattr(
-                    record, self.recipients_secondary.name, None)
-                if recipients_secondary:
-                    languagues.update(
-                        self._get_languages(recipients_secondary))
-                    cc = self._get_addresses(recipients_secondary)
-            if not cc and self.fallback_recipients_secondary:
-                languagues.update(
-                    self._get_languages(self.fallback_recipients_secondary))
-                cc = self._get_addresses(self.fallback_recipients_secondary)
-
-            bcc = []
-            if self.recipients_hidden:
-                recipients_hidden = getattr(
-                    record, self.recipients_hidden.name, None)
-                if recipients_hidden:
-                    languagues.update(self._get_languages(recipients_hidden))
-                    bcc = self._get_addresses(recipients_hidden)
-            if not bcc and self.fallback_recipients_hidden:
-                languagues.update(
-                    self._get_languages(self.fallback_recipients_hidden))
-                bcc = self._get_addresses(self.fallback_recipients_hidden)
-
-            to_addrs = [e for _, e in getaddresses(to + cc + bcc)]
+            to, to_languages = self._get_to(record)
+            cc, cc_languages = self._get_cc(record)
+            bcc, bcc_languages = self._get_bcc(record)
+            languages = to_languages | cc_languages | bcc_languages
+            to_addrs = [
+                convert_ascii_email(e) for _, e in getaddresses(to + cc + bcc)]
             if to_addrs:
-                msg = self.get_email(record, from_, to, cc, bcc, languagues)
+                msg, title = self.get_email(
+                    record, from_, to, cc, bcc, languages)
                 sendmail_transactional(
                     from_, to_addrs, msg, datamanager=datamanager)
-                logs.append(self.get_log(
-                        record, trigger, msg, bcc=', '.join(bcc)))
-        if logs:
-            Log.create(logs)
+                emails.append(Email(
+                        recipients=', '.join(to),
+                        recipients_secondary=', '.join(cc),
+                        recipients_hidden=', '.join(bcc),
+                        addresses=[{'address': a} for a in to_addrs],
+                        subject=title,
+                        resource=record,
+                        notification_email=trigger.notification_email,
+                        notification_trigger=trigger))
+        Email.save(emails)
+
+    def _get_recipients(self, record, name):
+        if name == 'id':
+            return record
+        else:
+            return getattr(record, name, None)
+
+    def _get_to(self, record):
+        to = []
+        languages = set()
+        if self.recipients:
+            recipients = self._get_recipients(record, self.recipients.name)
+            if recipients:
+                languages.update(self._get_languages(recipients))
+                to = self._get_addresses(recipients)
+        if not to and self.fallback_recipients:
+            languages.update(
+                self._get_languages(self.fallback_recipients))
+            to = self._get_addresses(self.fallback_recipients)
+        return to, languages
+
+    def _get_cc(self, record):
+        cc = []
+        languages = set()
+        if self.recipients_secondary:
+            recipients_secondary = self._get_recipients(
+                record, self.recipients_secondary.name)
+            if recipients_secondary:
+                languages.update(
+                    self._get_languages(recipients_secondary))
+                cc = self._get_addresses(recipients_secondary)
+        if not cc and self.fallback_recipients_secondary:
+            languages.update(
+                self._get_languages(self.fallback_recipients_secondary))
+            cc = self._get_addresses(self.fallback_recipients_secondary)
+        return cc, languages
+
+    def _get_bcc(self, record):
+        bcc = []
+        languages = set()
+        if self.recipients_hidden:
+            recipients_hidden = self._get_recipients(
+                record, self.recipients_hidden.name)
+            if recipients_hidden:
+                languages.update(self._get_languages(recipients_hidden))
+                bcc = self._get_addresses(recipients_hidden)
+        if not bcc and self.fallback_recipients_hidden:
+            languages.update(
+                self._get_languages(self.fallback_recipients_hidden))
+            bcc = self._get_addresses(self.fallback_recipients_hidden)
+        return bcc, languages
 
     @classmethod
-    def validate(cls, notifications):
-        super().validate(notifications)
-        for notification in notifications:
-            notification.check_subject()
+    def validate_fields(cls, notifications, field_names):
+        super().validate_fields(notifications, field_names)
+        cls.check_subject(notifications, field_names)
 
-    def check_subject(self):
-        if not self.subject:
+    @classmethod
+    def check_subject(cls, notifications, field_names=None):
+        if field_names and 'subject' not in field_names:
             return
-        try:
-            TextTemplate(self.subject)
-        except Exception as exception:
-            raise TemplateError(
-                gettext('notification_email.'
-                    'msg_notification_invalid_subject',
-                    notification=self.rec_name,
-                    exception=exception)) from exception
+        for notification in notifications:
+            if not notification.subject:
+                continue
+            try:
+                TextTemplate(notification.subject)
+            except Exception as exception:
+                raise TemplateError(gettext(
+                        'notification_email.'
+                        'msg_notification_invalid_subject',
+                        notification=notification.rec_name,
+                        exception=exception)) from exception
 
 
 class EmailAttachment(ModelSQL):
@@ -301,14 +338,12 @@ class EmailAttachment(ModelSQL):
     __name__ = 'notification.email.attachment'
 
     notification = fields.Many2One(
-        'notification.email', "Notification",
-        required=True, select=True)
+        'notification.email', "Notification", required=True)
     report = fields.Many2One(
         'ir.action.report', "Report", required=True,
         domain=[
             ('model', '=', Eval('model')),
-            ],
-        depends=['model'])
+            ])
 
     model = fields.Function(fields.Char("Model"), 'get_model')
 
@@ -340,66 +375,3 @@ class EmailAttachment(ModelSQL):
             'Content-Disposition', 'attachment',
             filename=('utf-8', language, name))
         return msg
-
-
-class EmailLog(ResourceAccessMixin, ModelSQL, ModelView):
-    "Notification Email Log"
-    __name__ = 'notification.email.log'
-    date = fields.Function(fields.DateTime('Date'), 'get_date')
-    recipients = fields.Char("Recipients")
-    recipients_secondary = fields.Char("Secondary Recipients")
-    recipients_hidden = fields.Char("Hidden Recipients")
-    notification = fields.Many2One(
-        'notification.email', "Notification",
-        required=True, ondelete='RESTRICT')
-    trigger = fields.Many2One('ir.trigger', "Trigger")
-
-    @classmethod
-    def __setup__(cls):
-        super().__setup__()
-        cls._order = [
-            ('create_date', 'DESC'),
-            ('id', 'DESC'),
-            ]
-
-    @classmethod
-    def __register__(cls, module_name):
-        pool = Pool()
-        Model = pool.get('ir.model')
-        Trigger = pool.get('ir.trigger')
-        model = Model.__table__()
-        trigger = Trigger.__table__()
-        table = cls.__table__()
-        super().__register__(module_name)
-
-        table_h = cls.__table_handler__(module_name)
-        cursor = Transaction().connection.cursor()
-
-        # Migration from 5.6:
-        # fill notification and resource
-        # remove required on trigger
-        notification = trigger.select(
-            trigger.notification_email,
-            where=trigger.id == table.trigger)
-        resource = (trigger
-            .join(model, condition=trigger.model == model.id)
-            .select(
-                Concat(model.model, ',-1'),
-                where=trigger.id == table.trigger))
-        cursor.execute(*table.update(
-                [table.notification, table.resource],
-                [notification, resource],
-                where=(table.trigger != Null) & (table.resource == Null)))
-        table_h.not_null_action('trigger', 'remove')
-
-    def get_date(self, name):
-        return self.create_date.replace(microsecond=0)
-
-    @classmethod
-    def search_date(cls, name, clause):
-        return [('create_date',) + tuple(clause[1:])]
-
-    @staticmethod
-    def order_date(tables):
-        table, _ = tables[None]
-        return [table.create_date]
